@@ -1,177 +1,10 @@
 #include "pairing.h"
-#include "fieldi.h"
-
-void cc_miller_vertical(element_t v, point_t Q, point_t Z)
-{
-    element_t t0;
-
-    if (point_is_inf(Z)) {
-	return;
-    }
-    element_init(t0, v->field);
-    element_sub(t0, Q->x, Z->x);
-    element_mul(v, v, t0);
-    element_clear(t0);
-
-    return;
-}
-
-void cc_miller_tangent(element_t v, point_t Q, point_t Z)
-{
-    //a = -slope_tangent(Z.x, Z.y);
-    //b = 1;
-    //c = -(Z.y + a * Z.x);
-    //but we multiply by 2*Z.y to avoid division
-
-    //a = -Zx * (Zx + Zx + Zx + twicea_2) - a_4;
-    //Common curves: a2 = 0 (and cc->a is a_4), so
-    //a = -(Zx (Zx + Zx + Zx) + cc->a)
-    //b = 2 * Zy
-    //c = -(2 Zy^2 + a Zx);
-    element_t a, b, c;
-    common_curve_ptr cc = Z->curve->data;
-    element_t t0;
-
-    if (point_is_inf(Z)) {
-	return;
-    }
-
-    if (element_is0(Z->y)) {
-	//order 2
-	cc_miller_vertical(v, Q, Z);
-	return;
-    }
-
-    element_init(a, v->field);
-    element_init(b, v->field);
-    element_init(c, v->field);
-    element_init(t0, v->field);
-
-    element_add(a, Z->x, Z->x);
-    element_add(a, a, Z->x);
-    element_mul(a, a, Z->x);
-    element_add(a, a, cc->a);
-    element_neg(a, a);
-
-    element_add(b, Z->y, Z->y);
-
-    element_mul(t0, b, Z->y);
-    element_mul(c, a, Z->x);
-    element_add(c, c, t0);
-    element_neg(c, c);
-
-    element_mul(a, a, Q->x);
-    element_mul(b, b, Q->y);
-    element_add(c, c, a);
-    element_add(c, c, b);
-    element_mul(v, v, c);
-
-    element_clear(a);
-    element_clear(b);
-    element_clear(c);
-    element_clear(t0);
-}
-
-void cc_miller_line(element_t v, point_t Q, point_t A, point_t B)
-{
-    //a = -(B.y - A.y) / (B.x - A.x);
-    //b = 1;
-    //c = -(A.y + a * A.x);
-    //but we'll multiply by B.x - A.x to avoid division
-    element_t a, b, c;
-    element_t t0;
-
-    //we assume B is never O
-    if (point_is_inf(A)) {
-	cc_miller_vertical(v, Q, A);
-	return;
-    }
-
-    if (!element_cmp(A->x, B->x)) {
-	if (!element_cmp(A->y, B->y)) {
-	    cc_miller_tangent(v, Q, A);
-	    return;
-	}
-	cc_miller_vertical(v, Q, A);
-	return;
-    }
-
-    element_init(a, v->field);
-    element_init(b, v->field);
-    element_init(c, v->field);
-    element_init(t0, v->field);
-
-    element_sub(b, B->x, A->x);
-    element_sub(a, A->y, B->y);
-    element_mul(t0, b, A->y);
-    element_mul(c, a, A->x);
-    element_add(c, c, t0);
-    element_neg(c, c);
-
-    element_mul(a, a, Q->x);
-    element_mul(b, b, Q->y);
-    element_add(c, c, a);
-    element_add(c, c, b);
-    element_mul(v, v, c);
-
-    element_clear(a);
-    element_clear(b);
-    element_clear(c);
-    element_clear(t0);
-}
+#include "solinas.h"
 
 //assumes P is in the base field, Q in some field extension,
 //and will be used in Tate pairing computation
-void old_cc_miller(element_t res, mpz_t q, point_t Pbase, point_t Q,
-	fieldmap mapbase)
-//doesn't use the fact that P lies in the base field
-//(we don't need Q until the last minute in miller_*()'s)
-{
-    //collate divisions
-    int m;
-    element_t v, vd;
-    point_t P;
-    point_t Z;
-
-    element_init(v, res->field);
-    element_init(vd, res->field);
-    point_init(Z, Q->curve);
-    point_init(P, Q->curve);
-
-    point_map(P, mapbase, Pbase);
-    point_set(Z, P);
-
-    element_set1(v);
-    element_set1(vd);
-    m = mpz_sizeinbase(q, 2) - 2;
-
-    while(m >= 0) {
-	element_mul(v, v, v);
-	element_mul(vd, vd, vd);
-	cc_miller_tangent(v, Q, Z);
-	point_double(Z, Z);
-	cc_miller_vertical(vd, Q, Z);
-	if (mpz_tstbit(q, m)) {
-	    cc_miller_line(v, Q, Z, P);
-	    point_add(Z, Z, P);
-	    cc_miller_vertical(vd, Q, Z);
-	}
-	m--;
-    }
-
-    element_invert(vd, vd);
-    element_mul(res, v, vd);
-
-    element_clear(v);
-    element_clear(vd);
-    point_clear(Z);
-    point_clear(P);
-}
-
-//assumes P is in the base field, Q in some field extension,
-//and will be used in Tate pairing computation
-void cc_miller(element_t res, mpz_t q, point_t P, point_t Q,
-	fieldmap mapbase)
+static void cc_miller(element_t res, mpz_t q, point_t P,
+	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
 {
     //collate divisions
     int m;
@@ -188,7 +21,7 @@ void cc_miller(element_t res, mpz_t q, point_t P, point_t Q,
 	    return;
 	}
 	mapbase(e0, Z->x);
-	element_sub(e0, Q->x, e0);
+	element_sub(e0, Qx, e0);
 	element_mul(e, e, e0);
     }
 
@@ -230,11 +63,11 @@ void cc_miller(element_t res, mpz_t q, point_t P, point_t Q,
 	element_add(c, c, t0);
 	element_neg(c, c);
 
-	//TODO: implement poly_mul_constant?
+	//TODO: use poly_mul_constant?
 	mapbase(e0, a);
-	element_mul(e0, e0, Q->x);
+	element_mul(e0, e0, Qx);
 	mapbase(e1, b);
-	element_mul(e1, e1, Q->y);
+	element_mul(e1, e1, Qy);
 	element_add(e0, e0, e1);
 	mapbase(e1, c);
 	element_add(e0, e0, e1);
@@ -276,9 +109,9 @@ void cc_miller(element_t res, mpz_t q, point_t P, point_t Q,
 	element_neg(c, c);
 
 	mapbase(e0, a);
-	element_mul(e0, e0, Q->x);
+	element_mul(e0, e0, Qx);
 	mapbase(e1, b);
-	element_mul(e1, e1, Q->y);
+	element_mul(e1, e1, Qy);
 	element_add(e0, e0, e1);
 	mapbase(e1, c);
 	element_add(e0, e0, e1);
@@ -334,11 +167,12 @@ static void cc_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	pairing_t pairing)
 {
     mnt_pairing_data_ptr p = pairing->data;
-    cc_miller(out, pairing->r, in1->data, in2->data, p->mapbase);
+    point_ptr Q = in2->data;
+    cc_miller(out, pairing->r, in1->data, Q->x, Q->y, p->mapbase);
     element_pow(out, out, p->tateexp);
 }
 
-static int cc_is_almost_coddh(element_ptr a, element_ptr b,
+static int cc_is_almost_coddh_odd_k(element_ptr a, element_ptr b,
 	element_ptr c, element_ptr d,
 	pairing_t pairing)
 {
@@ -349,8 +183,10 @@ static int cc_is_almost_coddh(element_ptr a, element_ptr b,
     element_init(t1, pairing->GT);
     element_init(t2, pairing->GT);
     mnt_pairing_data_ptr p = pairing->data;
-    cc_miller(t0, pairing->r, a->data, d->data, p->mapbase);
-    cc_miller(t1, pairing->r, b->data, c->data, p->mapbase);
+    point_ptr Pc = c->data;
+    point_ptr Pd = d->data;
+    cc_miller(t0, pairing->r, a->data, Pd->x, Pd->y, p->mapbase);
+    cc_miller(t1, pairing->r, b->data, Pc->x, Pc->y, p->mapbase);
     element_pow(t0, t0, p->tateexp);
     element_pow(t1, t1, p->tateexp);
     element_mul(t2, t0, t1);
@@ -391,7 +227,7 @@ static void trace(element_ptr out, element_ptr in, pairing_ptr pairing)
     point_clear(q);
 }
 
-void pairing_init_cc_param(pairing_t pairing, cc_param_t param)
+static void pairing_init_cc_param_odd_k(pairing_t pairing, cc_param_t param)
 {
     mnt_pairing_data_ptr p;
     element_t a, b;
@@ -402,7 +238,7 @@ void pairing_init_cc_param(pairing_t pairing, cc_param_t param)
     mpz_set(pairing->r, param->r);
     field_init_fp(pairing->Zr, pairing->r);
     pairing->map = cc_pairing;
-    pairing->is_almost_coddh = cc_is_almost_coddh;
+    pairing->is_almost_coddh = cc_is_almost_coddh_odd_k;
 
     p =	pairing->data = malloc(sizeof(mnt_pairing_data_t));
     field_init_fp(p->Fq, param->q);
@@ -426,7 +262,7 @@ void pairing_init_cc_param(pairing_t pairing, cc_param_t param)
 
     p->mapbase = ((polymod_field_data_ptr) p->Fqk->data)->mapbase;
 
-    cc_switch_field(p->Eqk, p->Eq, p->Fqk, p->mapbase);
+    cc_init_map_curve(p->Eqk, p->Eq, p->Fqk, p->mapbase);
 
     pairing->G1 = malloc(sizeof(field_t));
     pairing->G2 = malloc(sizeof(field_t));
@@ -444,23 +280,324 @@ void pairing_init_cc_param(pairing_t pairing, cc_param_t param)
     element_clear(b);
 }
 
+static void cc_miller_no_denom(element_t res, mpz_t q, point_t P,
+	element_ptr Qx, element_ptr Qy, fieldmap mapbase)
+{
+    //collate divisions
+    int m;
+    element_t v, vd;
+    point_t Z;
+    element_t a, b, c;
+    common_curve_ptr cc = P->curve->data;
+    element_t t0;
+    element_t e0, e1;
+
+    void do_vertical(element_t e)
+    {
+	if (point_is_inf(Z)) {
+	    return;
+	}
+	mapbase(e0, Z->x);
+	element_sub(e0, Qx, e0);
+	element_mul(e, e, e0);
+    }
+
+    void do_tangent(element_t e)
+    {
+	//a = -slope_tangent(Z.x, Z.y);
+	//b = 1;
+	//c = -(Z.y + a * Z.x);
+	//but we multiply by 2*Z.y to avoid division
+
+	//a = -Zx * (Zx + Zx + Zx + twicea_2) - a_4;
+	//Common curves: a2 = 0 (and cc->a is a_4), so
+	//a = -(Zx (Zx + Zx + Zx) + cc->a)
+	//b = 2 * Zy
+	//c = -(2 Zy^2 + a Zx);
+	element_ptr Zx = Z->x;
+	element_ptr Zy = Z->y;
+
+	if (point_is_inf(Z)) {
+	    return;
+	}
+
+	if (element_is0(Zy)) {
+	    //order 2
+	    do_vertical(e);
+	    return;
+	}
+
+	element_add(a, Zx, Zx);
+	element_add(a, a, Zx);
+	element_mul(a, a, Zx);
+	element_add(a, a, cc->a);
+	element_neg(a, a);
+
+	element_add(b, Zy, Zy);
+
+	element_mul(t0, b, Zy);
+	element_mul(c, a, Zx);
+	element_add(c, c, t0);
+	element_neg(c, c);
+
+	//TODO: use poly_mul_constant?
+	mapbase(e0, a);
+	element_mul(e0, e0, Qx);
+	mapbase(e1, b);
+	element_mul(e1, e1, Qy);
+	element_add(e0, e0, e1);
+	mapbase(e1, c);
+	element_add(e0, e0, e1);
+	element_mul(e, e, e0);
+    }
+
+    void do_line(element_ptr e)
+    {
+	//a = -(B.y - A.y) / (B.x - A.x);
+	//b = 1;
+	//c = -(A.y + a * A.x);
+	//but we'll multiply by B.x - A.x to avoid division
+
+	element_ptr Ax = Z->x;
+	element_ptr Ay = Z->y;
+	element_ptr Bx = P->x;
+	element_ptr By = P->y;
+
+	//we assume B is never O
+	if (point_is_inf(Z)) {
+	    do_vertical(e);
+	    return;
+	}
+
+	if (!element_cmp(Ax, Bx)) {
+	    if (!element_cmp(Ay, By)) {
+		do_tangent(e);
+		return;
+	    }
+	    do_vertical(e);
+	    return;
+	}
+
+	element_sub(b, Bx, Ax);
+	element_sub(a, Ay, By);
+	element_mul(t0, b, Ay);
+	element_mul(c, a, Ax);
+	element_add(c, c, t0);
+	element_neg(c, c);
+
+	mapbase(e0, a);
+	element_mul(e0, e0, Qx);
+	mapbase(e1, b);
+	element_mul(e1, e1, Qy);
+	element_add(e0, e0, e1);
+	mapbase(e1, c);
+	element_add(e0, e0, e1);
+	element_mul(e, e, e0);
+    }
+
+    element_init(a, P->curve->field);
+    element_init(b, P->curve->field);
+    element_init(c, P->curve->field);
+    element_init(t0, P->curve->field);
+    element_init(e0, res->field);
+    element_init(e1, res->field);
+
+    element_init(v, res->field);
+    element_init(vd, res->field);
+    point_init(Z, P->curve);
+
+    point_set(Z, P);
+
+    element_set1(v);
+    element_set1(vd);
+    m = mpz_sizeinbase(q, 2) - 2;
+
+    while(m >= 0) {
+	element_mul(v, v, v);
+	do_tangent(v);
+	point_double(Z, Z);
+	if (mpz_tstbit(q, m)) {
+	    do_line(v);
+	    point_add(Z, Z, P);
+	}
+	m--;
+    }
+
+    element_set(res, v);
+
+    element_clear(v);
+    element_clear(vd);
+    point_clear(Z);
+    element_clear(a);
+    element_clear(b);
+    element_clear(c);
+    element_clear(t0);
+    element_clear(e0);
+    element_clear(e1);
+}
+
+static void cc_pairing_even_k(element_ptr out, element_ptr in1, element_ptr in2,
+	pairing_t pairing)
+{
+    point_ptr Qbase = in2->data;
+    element_t x, y;
+    even_mnt_pairing_data_ptr p = pairing->data;
+
+    element_init(x, out->field);
+    element_init(y, out->field);
+    //map from twist: (x, y) --> (v^-1 x, v^-(3/2) y)
+    //where v is the quadratic nonresidue used to construct the twist
+    element_mul(fi_re(x), Qbase->x, p->nqrinv);
+    //v^-3/2 = v^-2 * v^1/2
+    element_mul(fi_im(y), Qbase->y, p->nqrinv2);
+    cc_miller_no_denom(out, pairing->r, in1->data, x, y, p->mapbase);
+    element_pow(out, out, p->tateexp);
+}
+
+static int cc_is_almost_coddh_even_k(element_ptr a, element_ptr b,
+	element_ptr c, element_ptr d,
+	pairing_t pairing)
+{
+    int res = 0;
+    element_t t0, t1, t2;
+    element_t cx, cy;
+    element_t dx, dy;
+    even_mnt_pairing_data_ptr p = pairing->data;
+
+    element_init(cx, p->Fqk);
+    element_init(cy, p->Fqk);
+    element_init(dx, p->Fqk);
+    element_init(dy, p->Fqk);
+
+    element_init(t0, pairing->GT);
+    element_init(t1, pairing->GT);
+    element_init(t2, pairing->GT);
+    point_ptr Pc = c->data;
+    point_ptr Pd = d->data;
+    //map from twist: (x, y) --> (v^-1 x, v^-(3/2) y)
+    //where v is the quadratic nonresidue used to construct the twist
+    element_mul(fi_re(cx), Pc->x, p->nqrinv);
+    element_mul(fi_re(dx), Pd->x, p->nqrinv);
+    //v^-3/2 = v^-2 * v^1/2
+    element_mul(fi_im(cy), Pc->y, p->nqrinv2);
+    element_mul(fi_im(dy), Pd->y, p->nqrinv2);
+
+    cc_miller_no_denom(t0, pairing->r, a->data, dx, dy, p->mapbase);
+    cc_miller(t1, pairing->r, b->data, cx, cy, p->mapbase);
+    element_pow(t0, t0, p->tateexp);
+    element_pow(t1, t1, p->tateexp);
+    element_mul(t2, t0, t1);
+    if (element_is1(t2)) {
+	//g, g^x, h, h^-x case
+	res = 1;
+    } else {
+	element_invert(t1, t1);
+	element_mul(t2, t0, t1);
+	if (element_is1(t2)) {
+	    //g, g^x, h, h^x case
+	    res = 1;
+	}
+    }
+    element_clear(t0);
+    element_clear(t1);
+    element_clear(t2);
+    return res;
+}
+
+static void Fq_to_Fqd_to_Fqk(element_t out, element_t in)
+{
+    element_t e0;
+    fq_field_data_ptr p = out->field->data;
+    element_init(e0, p->field);
+    element_field_to_poly(e0, in);
+    element_field_to_quadratic(out, e0);
+    element_clear(e0);
+}
+
+static void pairing_init_cc_param_even_k(pairing_t pairing, cc_param_t param)
+{
+    even_mnt_pairing_data_ptr p;
+    element_t a, b;
+    element_t irred;
+    mpz_t z;
+    int d = param->k / 2;
+
+    mpz_init(pairing->r);
+    mpz_set(pairing->r, param->r);
+    field_init_fp(pairing->Zr, pairing->r);
+    pairing->map = cc_pairing_even_k;
+    pairing->is_almost_coddh = cc_is_almost_coddh_even_k;
+
+    p =	pairing->data = malloc(sizeof(even_mnt_pairing_data_t));
+    field_init_fp(p->Fq, param->q);
+    element_init(a, p->Fq);
+    element_init(b, p->Fq);
+    element_set_mpz(a, param->a);
+    element_set_mpz(b, param->b);
+    curve_init_cc_ab(p->Eq, a, b);
+
+    field_init_poly(p->Fqx, p->Fq);
+    element_init(irred, p->Fqx);
+    do {
+	poly_random_monic(irred, d);
+    } while (!poly_is_irred(irred));
+    field_init_polymod(p->Fqd, irred);
+    element_clear(irred);
+    field_init_quadratic(p->Fqk, p->Fqd);
+
+    mpz_init(p->tateexp);
+    mpz_sub_ui(p->tateexp, p->Fqk->order, 1);
+    mpz_divexact(p->tateexp, p->tateexp, pairing->r);
+
+    p->mapbase = Fq_to_Fqd_to_Fqk;
+
+    cc_init_map_curve(p->Etwist, p->Eq, p->Fqd, element_field_to_poly);
+    twist_curve(p->Etwist);
+    element_init(p->nqrinv, p->Fqd);
+    element_invert(p->nqrinv, field_get_nqr(p->Fqd));
+    element_init(p->nqrinv2, p->Fqd);
+    //TODO: element_square
+    element_mul(p->nqrinv2, p->nqrinv, p->nqrinv);
+
+    pairing->G1 = malloc(sizeof(field_t));
+    pairing->G2 = malloc(sizeof(field_t));
+
+    field_init_curve_group(pairing->G1, p->Eq, param->h);
+    mpz_init(z);
+    mpz_set_si(z, 1);
+    field_init_curve_group(pairing->G2, p->Etwist, z);
+    mpz_clear(z);
+    p->k = param->k;
+    pairing->GT = p->Fqk;
+    //pairing->phi = trace;
+
+    element_clear(a);
+    element_clear(b);
+}
+
+void pairing_init_cc_param(pairing_t pairing, cc_param_t param)
+{
+    if (param->k % 2) pairing_init_cc_param_odd_k(pairing, param);
+    else pairing_init_cc_param_even_k(pairing, param);
+}
+
 static void phi_identity(element_ptr out, element_ptr in, pairing_ptr pairing)
 {
     element_set(out, in);
 }
 
-//TODO: use Miller algorithm that's been specialized for Solinas primes
 static void solinas_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	pairing_t pairing)
 //in1, in2 are from E(F_q), out from F_q^2
 {
     solinas_pairing_data_ptr p = pairing->data;
-    point_ptr Qbase = in2->data;
     point_t V, V1;
     element_t f, f0, f1;
     element_t a, b, c;
     element_t e0;
     int i, n;
+    element_ptr Qx = ((point_ptr) in2->data)->x;
+    element_ptr Qy = ((point_ptr) in2->data)->y;
 
     void do_tangent() {
 	//a = -slope_tangent(V.x, V.y);
@@ -486,12 +623,12 @@ static void solinas_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	element_add(c, c, e0);
 	element_neg(c, c);
 
-	//we'll map Qbase via (x,y) --> (-x, iy)
-	//hence a Q.x + c = -a Qbase.x + c is real while
-	//(b Q.y) = b Qbase.y i is purely imaginary.
-	element_mul(a, a, Qbase->x);
+	//we'll map Q via (x,y) --> (-x, iy)
+	//hence a Qx + c = -a Qx + c is real while
+	//(b Qy) = b Qy i is purely imaginary.
+	element_mul(a, a, Qx);
 	element_sub(fi_re(f0), c, a);
-	element_mul(fi_im(f0), b, Qbase->y);
+	element_mul(fi_im(f0), b, Qy);
 	element_mul(f, f, f0);
     }
 
@@ -510,36 +647,37 @@ static void solinas_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	element_add(c, c, e0);
 	element_neg(c, c);
 
-	//we'll map Qbase via (x,y) --> (-x, iy)
-	//hence a Q.x + c = -a Qbase.x + c is real while
-	//(b Q.y) = b Qbase.y i is purely imaginary.
-	element_mul(a, a, Qbase->x);
+	//we'll map Q via (x,y) --> (-x, iy)
+	//hence a Qx + c = -a Qx + c is real while
+	//(b Qy) = b Qy i is purely imaginary.
+	element_mul(a, a, Qx);
 	element_sub(fi_re(f0), c, a);
-	element_mul(fi_im(f0), b, Qbase->y);
+	element_mul(fi_im(f0), b, Qy);
 	element_mul(f, f, f0);
     }
 
    if (0) {
 	//check against standard Miller's alogrithm
-	point_t Q;
-	fi_data_ptr fdp;
+	fq_data_ptr fdp;
+	element_t x, y;
 
-	point_init(Q, p->Eq2);
-	Q->inf_flag = 0;
-	fdp = Q->x->data;
-	element_neg(fdp->x, Qbase->x);
+	element_init(x, p->Fq2);
+	element_init(y, p->Fq2);
+	fdp = x->data;
+	element_neg(fdp->x, Qx);
 	element_set0(fdp->y);
-	fdp = Q->y->data;
-	element_set(fdp->y, Qbase->y);
+	fdp = y->data;
+	element_set(fdp->y, Qy);
 	element_set0(fdp->x);
 
-	cc_miller(out, pairing->r, in1->data, Q, element_field_to_fieldi);
+	cc_miller(out, pairing->r, in1->data, x, y, element_field_to_quadratic);
 	element_pow(out, out, p->tateexp);
 	printf("check: ");
 	element_out_str(stdout, 0, out);
 	printf("\n");
 
-	point_clear(Q);
+	element_clear(x);
+	element_clear(y);
     }
 
     point_init(V, p->Eq);
@@ -563,10 +701,8 @@ static void solinas_pairing(element_ptr out, element_ptr in1, element_ptr in2,
 	point_double(V, V);
     }
     if (p->sign1 < 0) {
-	//TODO: do_vert() or something too
 	point_neg(V1, V);
-	element_invert(f0, f);
-printf("negative sign1 not implemented yet!\n");
+	element_invert(f1, f);
     } else {
 	point_set(V1, V);
 	element_set(f1, f);
@@ -581,12 +717,6 @@ printf("negative sign1 not implemented yet!\n");
     element_mul(f, f, f1);
     do_line(V, V1);
 
-    if (p->sign0 > 0) {
-	//do something
-printf("positive sign0 not implemented yet!\n");
-    }
-    //otherwise we do nothing since f_r = f_{r-1}
-
     //Tate exponentiation
     //simpler but slower:
     //element_pow(out, f, p->tateexp);
@@ -616,12 +746,11 @@ void pairing_init_solinas_param(pairing_t pairing, solinas_param_t param)
     p->exp2 = param->exp2;
     p->exp1 = param->exp1;
     p->sign1 = param->sign1;
-    p->sign0 = param->sign0;
     mpz_init(pairing->r);
     mpz_set(pairing->r, param->r);
     field_init_fp(pairing->Zr, pairing->r);
     pairing->map = solinas_pairing;
-    pairing->is_almost_coddh = cc_is_almost_coddh;
+    //pairing->is_almost_coddh = cc_is_almost_coddh;
 
     field_init_fp(p->Fq, param->q);
     element_init(a, p->Fq);
@@ -633,7 +762,6 @@ void pairing_init_solinas_param(pairing_t pairing, solinas_param_t param)
     element_clear(b);
 
     field_init_fi(p->Fq2, p->Fq);
-    cc_switch_field(p->Eq2, p->Eq, p->Fq2, element_field_to_fieldi);
 
     mpz_init(p->tateexp);
     mpz_sub_ui(p->tateexp, p->Fq2->order, 1);
