@@ -37,6 +37,8 @@ int main()
     element_out_str(stdout, 0, c);
     printf("\n");
 
+    { int i, n = 10;
+	for (i=0; i<n; i++) {
     for (;;) {
 	element_random(a);
 	printf("new a = ");
@@ -54,6 +56,8 @@ int main()
     printf("sqrt(a) * sqrt(a) = ");
     element_out_str(stdout, 0, c);
     printf("\n");
+	}
+    }
 
     return 0;
 }
